@@ -5,7 +5,7 @@ Queries may be by point, by range overlap, or by range envelopment.
 Core logic.
 
 Copyright 2013-2014 Chaim-Leib Halbert
-Modifications Copyright 2014 Konstantin Tretrakov
+Modifications Copyright 2014 Konstantin Tretyakov
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -19,10 +19,11 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """
-from __future__ import absolute_import
 from .interval import Interval
 from .node import Node
 from numbers import Number
+import collections
+from warnings import warn
 
 try:
     xrange  # Python 2?
@@ -31,7 +32,7 @@ except NameError:
 
 
 # noinspection PyBroadException
-class IntervalTree(object):
+class IntervalTree(collections.MutableSet):
     """
     A binary lookup tree of intervals.
     The intervals contained in the tree are represented using ``Interval(a, b, data)`` objects.
@@ -70,7 +71,7 @@ class IntervalTree(object):
         >>> tree.addi(19.9, 20)
         >>> tree
         IntervalTree([Interval(0, 1, 'data'), Interval(10, 20), Interval(19.9, 20)])
-        >>> tree.extend([Interval(19.9, 20.1), Interval(20.1, 30)])
+        >>> tree.update([Interval(19.9, 20.1), Interval(20.1, 30)])
         >>> len(tree)
         5
 
@@ -116,9 +117,12 @@ class IntervalTree(object):
         >>> tree.remove_overlap(1.7, 1.8)
         >>> tree
         IntervalTree([Interval(0.5, 1.7)])
-        >>> tree.remove_overlap(1.6, 1.6) # Empty interval still works
+        >>> tree.remove_overlap(1.6, 1.6)  # Null interval does nothing
         >>> tree
-        IntervalTree()
+        IntervalTree([Interval(0.5, 1.7)])
+        >>> tree.remove_overlap(1.6, 1.5)  # Ditto
+        >>> tree
+        IntervalTree([Interval(0.5, 1.7)])
         
     Delete intervals, enveloped in the range::
     
@@ -142,8 +146,10 @@ class IntervalTree(object):
         >>> assert tree[-1.1]         == set([Interval(-1.1, 1.1)])
         >>> assert tree.search(1.1)   == set([Interval(-0.5, 1.5), Interval(0.5, 1.7)])   # Same as tree[1.1]
         >>> assert tree[-0.5:0.5]     == set([Interval(-0.5, 1.5), Interval(-1.1, 1.1)])  # Interval overlap query
-        >>> assert tree.search(1.5, 1.5) == set([Interval(0.5, 1.7)])                     # Same as tree[1.5:1.5]
-        >>> assert tree.search(1.7, 1.7) == set()
+        >>> assert tree.search(1.5, 1.5) == set()                                         # Same as tree[1.5:1.5]
+        >>> assert tree.search(1.5) == set([Interval(0.5, 1.7)])                          # Same as tree[1.5]
+
+        >>> assert tree.search(1.7, 1.8) == set()
 
     Envelop queries::
     
@@ -161,8 +167,8 @@ class IntervalTree(object):
         False
         >>> tree.overlaps(-1.1)
         True
-        >>> not tree.overlaps(1.7) # TODO: itree.overlaps(1.7) returns None, should return False
-        True
+        >>> tree.overlaps(1.7)
+        False
         >>> tree.overlaps(1.7, 1.8)
         False
         >>> tree.overlaps(-1.2, -1.1)
@@ -219,7 +225,15 @@ class IntervalTree(object):
         >>> IntervalTree([Interval(0, 1)]) == IntervalTree([Interval(0, 1, "x")])
         False
     """
-    
+    @classmethod
+    def from_tuples(cls, tups):
+        """
+        Create a new IntervalTree from an iterable of 2- or 3-tuples,
+         where the tuple lists begin, end, and optionally data.
+        """
+        ivs = [Interval(*t) for t in tups]
+        return IntervalTree(ivs)
+
     def __init__(self, intervals=None):
         """
         Set up a tree. If intervals is provided, add all the intervals 
@@ -239,7 +253,6 @@ class IntervalTree(object):
         self.boundary_table = {}
         for iv in self.all_intervals:
             self._add_boundaries(iv)
-        #self.verify()
 
     def copy(self):
         """
@@ -321,7 +334,7 @@ class IntervalTree(object):
         return self.add(Interval(begin, end, data))
     appendi = addi
     
-    def extend(self, intervals):
+    def update(self, intervals):
         """
         Given an iterable of intervals, add them to the tree.
         
@@ -330,7 +343,14 @@ class IntervalTree(object):
         """
         for iv in intervals:
             self.add(iv)
-    
+
+    def extend(self, intervals):
+        """
+        Deprecated: Replaced by update().
+        """
+        warn("IntervalTree.extend() has been deprecated. Consider using update() instead", DeprecationWarning)
+        self.update(intervals)
+
     def remove(self, interval):
         """
         Removes an interval from the tree, if present. If not, raises 
@@ -375,7 +395,74 @@ class IntervalTree(object):
         Completes in O(log n) time.
         """
         return self.discard(Interval(begin, end, data))
-    
+
+    def difference(self, other):
+        """
+        Returns a new tree, comprising all intervals in self but not
+        in other.
+        """
+        ivs = set()
+        for iv in self:
+            if iv not in other:
+                ivs.add(iv)
+        return IntervalTree(ivs)
+
+    def difference_update(self, other):
+        """
+        Removes all intervals in other from self.
+        """
+        for iv in other:
+            self.discard(iv)
+
+    def union(self, other):
+        """
+        Returns a new tree, comprising all intervals from self
+        and other.
+        """
+        return IntervalTree(set(self).union(other))
+
+    def intersection(self, other):
+        """
+        Returns a new tree of all intervals common to both self and
+        other.
+        """
+        ivs = set()
+        shorter, longer = sorted([self, other], key=len)
+        for iv in shorter:
+            if iv in longer:
+                ivs.add(iv)
+        return IntervalTree(ivs)
+
+    def intersection_update(self, other):
+        """
+        Removes intervals from self unless they also exist in other.
+        """
+        for iv in self:
+            if iv not in other:
+                self.remove(iv)
+
+    def symmetric_difference(self, other):
+        """
+        Return a tree with elements only in self or other but not
+        both.
+        """
+        if not isinstance(other, set): other = set(other)
+        me = set(self)
+        ivs = me  - other + (other - me)
+        return IntervalTree(ivs)
+
+    def symmetric_difference_update(self, other):
+        """
+        Throws out all intervals except those only in self or other,
+        not both.
+        """
+        other = set(other)
+        for iv in self:
+            if iv in other:
+                self.remove(iv)
+                other.remove(iv)
+        self.update(other)
+
     def remove_overlap(self, begin, end=None):
         """
         Removes all intervals overlapping the given point or range.
@@ -388,7 +475,7 @@ class IntervalTree(object):
         hitlist = self.search(begin, end)
         for iv in hitlist: 
             self.remove(iv)
-    
+
     def remove_envelop(self, begin, end):
         """
         Removes all intervals completely enveloped in the given range.
@@ -401,7 +488,62 @@ class IntervalTree(object):
         hitlist = self.search(begin, end, strict=True)
         for iv in hitlist:
             self.remove(iv)
-        
+
+    def chop(self, begin, end, datafunc=None):
+        """
+        Like remove_envelop(), but trims back Intervals hanging into
+        the chopped area so that nothing overlaps.
+        """
+        insertions = set()
+        begin_hits = [iv for iv in self[begin] if iv.begin < begin]
+        end_hits = [iv for iv in self[end] if iv.end > end]
+
+        if datafunc:
+            for iv in begin_hits:
+                insertions.add(Interval(iv.begin, begin, datafunc(iv, True)))
+            for iv in end_hits:
+                insertions.add(Interval(end, iv.end, datafunc(iv, False)))
+        else:
+            for iv in begin_hits:
+                insertions.add(Interval(iv.begin, begin, iv.data))
+            for iv in end_hits:
+                insertions.add(Interval(end, iv.end, iv.data))
+
+        self.remove_envelop(begin, end)
+        self.difference_update(begin_hits)
+        self.difference_update(end_hits)
+        self.update(insertions)
+
+    def slice(self, point, datafunc=None):
+        """
+        Split Intervals that overlap point into two new Intervals. if
+        specified, uses datafunc(interval, islower=True/False) to
+        set the data field of the new Intervals.
+        :param point: where to slice
+        :param datafunc(interval, isupper): callable returning a new
+        value for the interval's data field
+        """
+        hitlist = set(iv for iv in self[point] if iv.begin < point)
+        insertions = set()
+        if datafunc:
+            for iv in hitlist:
+                insertions.add(Interval(iv.begin, point, datafunc(iv, True)))
+                insertions.add(Interval(point, iv.end, datafunc(iv, False)))
+        else:
+            for iv in hitlist:
+                insertions.add(Interval(iv.begin, point, iv.data))
+                insertions.add(Interval(point, iv.end, iv.data))
+        self.difference_update(hitlist)
+        self.update(insertions)
+
+    def clear(self):
+        """
+        Empties the tree.
+
+        Completes in O(1) tine.
+        """
+        self.__init__()
+
     def find_nested(self):
         """
         Returns a dictionary mapping parent intervals to sets of 
@@ -418,7 +560,7 @@ class IntervalTree(object):
                     result[parent] = set()
                 result[parent].add(child)
                 
-        long_ivs = sorted(self.all_intervals, key=len, reverse=True)
+        long_ivs = sorted(self.all_intervals, key=Interval.length, reverse=True)
         for i, parent in enumerate(long_ivs):
             for child in long_ivs[i + 1:]:
                 add_if_nested()
@@ -449,7 +591,7 @@ class IntervalTree(object):
         """
         if self.is_empty():
             return False
-        return self.top_node.contains_point(p)
+        return bool(self.top_node.contains_point(p))
     
     def overlaps_range(self, begin, end):
         """
@@ -464,7 +606,6 @@ class IntervalTree(object):
             return False
         elif self.overlaps_point(begin):
             return True
-        # TODO: add support for open and closed intervals
         return any(
             self.overlaps_point(bound) 
             for bound in self.boundary_table 
@@ -532,6 +673,8 @@ class IntervalTree(object):
                 return self.search(iv.begin, iv.end, strict=strict)
             except:
                 return self.top_node.search_point(begin, set())
+        elif begin >= end:
+            return set()
         else:
             result = self.top_node.search_point(begin, set())
             
@@ -699,8 +842,15 @@ class IntervalTree(object):
         :rtype: set of Interval
         """
         try:
-            return self.search(index.start, index.stop)
-        except:
+            start, stop = index.start, index.stop
+            if start is None:
+                start = self.begin()
+                if stop is None:
+                    return set(self)
+            if stop is None:
+                stop = self.end()
+            return self.search(start, stop)
+        except AttributeError:
             return self.search(index)
     
     def __setitem__(self, index, value):
@@ -715,6 +865,12 @@ class IntervalTree(object):
         """
         self.addi(index.start, index.stop, value)
 
+    def __delitem__(self, point):
+        """
+        Delete all items overlapping point.
+        """
+        self.remove_overlap(point)
+
     def __contains__(self, item):
         """
         Returns whether item exists as an Interval in the tree.
